from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector


class CostAnalysisManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cost_analysis_conn: SpaceConnector = self.locator.get_connector(
            "SpaceConnector", service="cost_analysis"
        )

    def analyze_cost(self, params: dict) -> dict:
        return self.cost_analysis_conn.dispatch("Cost.analyze", params)



from typing import Union, Literal
from pydantic import BaseModel

__all__ = [
    "PrivateDataTableAddRequest",
    "PrivateDataTableTransformRequest",
    "PrivateDataTableUpdateRequest",
    "PrivateDataTableDeleteRequest",
    "PrivateDataTableLoadRequest",
    "PrivateDataTableGetRequest",
    "PrivateDataTableSearchQueryRequest",
]

Granularity = Literal["DAILY", "MONTHLY", "YEARLY"]


class PrivateDataTableAddRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    source_type: str
    options: dict
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateDataTableTransformRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    operator: str
    options: dict
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateDataTableUpdateRequest(BaseModel):
    data_table_id: str
    name: Union[str, None] = None
    options: Union[dict, None] = None
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateDataTableDeleteRequest(BaseModel):
    data_table_id: str
    user_id: str
    domain_id: str


class PrivateDataTableLoadRequest(BaseModel):
    data_table_id: str
    granularity: Granularity
    start: Union[str, None] = None
    end: Union[str, None] = None
    sort: Union[list, None] = None
    page: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateDataTableGetRequest(BaseModel):
    data_table_id: str
    user_id: str
    domain_id: str


class PrivateDataTableSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    widget_id: str
    data_table_id: Union[str, None] = None
    name: Union[str, None] = None
    data_type: Union[str, None] = None
    source_type: Union[str, None] = None
    operator: Union[str, None] = None
    user_id: str
    domain_id: str


class PrivateDataTableStatQueryRequest(BaseModel):
    query: dict
    user_id: str
    domain_id: str



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.private_dashboard_manager import PrivateDashboardManager
from spaceone.dashboard.manager.private_folder_manager import PrivateFolderManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.private_dashboard.request import *
from spaceone.dashboard.model.private_dashboard.response import *
from spaceone.dashboard.model.private_dashboard.database import PrivateDashboard

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateDashboardService(BaseService):
    resource = "PrivateDashboard"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_dashboard_mgr = PrivateDashboardManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def create(
        self, params: PrivateDashboardCreateRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Create private dashboard

        Args:
            params (dict): {
                'name': 'str',                  # required
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'dict',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'workspace_id': 'str',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        if params.folder_id:
            pri_folder_mgr = PrivateFolderManager()
            pri_folder_mgr.get_private_folder(
                params.folder_id, params.domain_id, params.user_id
            )

        if params.workspace_id:
            self.identity_mgr.check_workspace(params.workspace_id, params.domain_id)

        pri_dashboard_vo = self.pri_dashboard_mgr.create_private_dashboard(
            params.dict()
        )
        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateDashboardUpdateRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Update private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'name': 'str',
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'list',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        pri_dashboard_vo: PrivateDashboard = (
            self.pri_dashboard_mgr.get_private_dashboard(
                params.dashboard_id, params.domain_id, params.user_id
            )
        )

        if params.folder_id:
            pri_folder_mgr = PrivateFolderManager()
            pri_folder_mgr.get_private_folder(
                params.folder_id, params.domain_id, params.user_id
            )

        pri_dashboard_vo = self.pri_dashboard_mgr.update_private_dashboard_by_vo(
            params.dict(exclude_unset=True), pri_dashboard_vo
        )

        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDashboard.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateDashboardDeleteRequest) -> None:
        """Delete private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_dashboard_vo = self.pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id, params.domain_id, params.user_id
        )

        self.pri_dashboard_mgr.delete_private_dashboard_by_vo(pri_dashboard_vo)

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @convert_model
    def get(
        self, params: PrivateDashboardGetRequest
    ) -> Union[PrivateDashboardResponse, dict]:
        """Get private dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDashboardResponse:
        """

        pri_dashboard_vo = self.pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id, params.domain_id, params.user_id
        )

        return PrivateDashboardResponse(**pri_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @append_query_filter(
        ["dashboard_id", "name", "domain_id", "workspace_id", "user_id", "folder_id"]
    )
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def list(
        self, params: PrivateDashboardSearchQueryRequest
    ) -> Union[PrivateDashboardsResponse, dict]:
        """List private dashboards

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',
                'name': 'str',
                'folder_id': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'workspace_id': 'str',
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateDashboardsResponse:
        """

        query = params.query or {}
        pri_dashboard_vos, total_count = self.pri_dashboard_mgr.list_private_dashboards(
            query
        )
        pri_dashboards_info = [
            pri_dashboard_vo.to_dict() for pri_dashboard_vo in pri_dashboard_vos
        ]
        return PrivateDashboardsResponse(
            results=pri_dashboards_info, total_count=total_count
        )

    @transaction(
        permission="dashboard:PrivateDashboard.read",
        role_types=["USER"],
    )
    @append_query_filter(["domain_id", "user_id"])
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def stat(self, params: PrivateDashboardStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'user_id': 'str',                                       # injected from auth (required)
                'domain_id': 'str'                                      # injected from auth (required)
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pri_dashboard_mgr.stat_private_dashboards(query)



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.public_dashboard_manager import PublicDashboardManager
from spaceone.dashboard.manager.public_folder_manager import PublicFolderManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.public_dashboard.request import *
from spaceone.dashboard.model.public_dashboard.response import *
from spaceone.dashboard.model.public_dashboard.database import PublicDashboard

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PublicDashboardService(BaseService):
    resource = "PublicDashboard"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pub_dashboard_mgr = PublicDashboardManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def create(
        self, params: PublicDashboardCreateRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Create public dashboard

        Args:
            params (dict): {
                'name': 'str',                  # required
                'description': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'dict',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'resource_group': 'str',        # required
                'project_id': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PublicDashboardResponse:
        """

        if params.resource_group == "PROJECT":
            if not params.project_id:
                raise ERROR_REQUIRED_PARAMETER(key="project_id")

            project_info = self.identity_mgr.get_project(params.project_id)
            params.workspace_id = project_info["workspace_id"]
        elif params.resource_group == "WORKSPACE":
            if not params.workspace_id:
                raise ERROR_REQUIRED_PARAMETER(key="workspace_id")

            self.identity_mgr.check_workspace(params.workspace_id, params.domain_id)
            params.project_id = "*"
        else:
            params.workspace_id = "*"
            params.project_id = "*"

        if params.folder_id:
            pub_folder_mgr = PublicFolderManager()
            pub_folder_mgr.get_public_folder(
                params.folder_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
                params.resource_group,
            )

        pub_dashboard_vo = self.pub_dashboard_mgr.create_public_dashboard(params.dict())
        return PublicDashboardResponse(**pub_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def update(
        self, params: PublicDashboardUpdateRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Update public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'description': 'str',
                'name': 'str',
                'layouts': 'list',
                'vars': 'dict',
                'options': 'dict',
                'variables': 'dict',
                'variables_schema': 'list',
                'labels': 'list',
                'tags': 'dict',
                'folder_id': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if params.folder_id:
            pub_folder_mgr = PublicFolderManager()
            pub_folder_mgr.get_public_folder(
                params.folder_id,
                pub_dashboard_vo.domain_id,
                pub_dashboard_vo.workspace_id,
                params.user_projects,
                pub_dashboard_vo.resource_group,
            )

        pub_dashboard_vo = self.pub_dashboard_mgr.update_public_dashboard_by_vo(
            params.dict(exclude_unset=True), pub_dashboard_vo
        )

        return PublicDashboardResponse(**pub_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDashboard.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def delete(self, params: PublicDashboardDeleteRequest) -> None:
        """Delete public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        self.pub_dashboard_mgr.delete_public_dashboard_by_vo(pub_dashboard_vo)

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def get(
        self, params: PublicDashboardGetRequest
    ) -> Union[PublicDashboardResponse, dict]:
        """Get public dashboard

        Args:
            params (dict): {
                'dashboard_id': 'str',   # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicDashboardResponse:
        """

        pub_dashboard_vo: PublicDashboard = self.pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        return PublicDashboardResponse(**pub_dashboard_vo.to_dict())

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(
        [
            "dashboard_id",
            "name",
            "domain_id",
            "workspace_id",
            "project_id",
            "folder_id",
            "user_projects",
        ]
    )
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def list(
        self, params: PublicDashboardSearchQueryRequest
    ) -> Union[PublicDashboardsResponse, dict]:
        """List public dashboards

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',
                'name': 'str',
                'folder_id': 'str',
                'project_id': 'str',
                'workspace_id': 'str',                          # injected from auth
                'domain_id': 'str',                             # injected from auth (required)
                'user_projects': 'list',                        # injected from auth
            }

        Returns:
            PublicDashboardsResponse:
        """

        query = params.query or {}
        pub_dashboard_vos, total_count = self.pub_dashboard_mgr.list_public_dashboards(
            query
        )
        pub_dashboards_info = [
            pub_dashboard_vo.to_dict() for pub_dashboard_vo in pub_dashboard_vos
        ]
        return PublicDashboardsResponse(
            results=pub_dashboards_info, total_count=total_count
        )

    @transaction(
        permission="dashboard:PublicDashboard.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["domain_id", "workspace_id", "user_projects"])
    @append_keyword_filter(["dashboard_id", "name"])
    @convert_model
    def stat(self, params: PublicDashboardStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'workspace_id': 'str',                                  # injected from auth
                'domain_id': 'str'                                      # injected from auth (required)
                'user_projects': 'list',                                # injected from auth
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pub_dashboard_mgr.stat_public_dashboards(query)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PublicDashboard(MongoModel):
    dashboard_id = StringField(
        max_length=40, generate_id="public-dash", unique=True
    )
    name = StringField(max_length=100)
    description = StringField(default=None)
    version = StringField(max_length=40, default="2.0")
    layouts = ListField(default=None)
    vars = DictField(default=None)
    options = DictField(default=None)
    variables = DictField(default=None)
    variables_schema = DictField(default=None)
    labels = ListField(StringField())
    tags = DictField(default=None)
    resource_group = StringField(
        max_length=40, choices=("DOMAIN", "WORKSPACE", "PROJECT")
    )
    project_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "description",
            "layouts",
            "vars",
            "options",
            "variables",
            "variables_schema",
            "labels",
            "tags",
        ],
        "minimal_fields": [
            "dashboard_id",
            "name",
            "version",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
        "change_query_keys": {"user_projects": "project_id"},
        "ordering": ["name"],
        "indexes": [
            "name",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
    }



from spaceone.api.dashboard.v1 import (
    private_data_table_pb2,
    private_data_table_pb2_grpc,
)
from spaceone.core.pygrpc import BaseAPI
from spaceone.dashboard.service.private_data_table_service import (
    PrivateDataTableService,
)


class PrivateDataTable(BaseAPI, private_data_table_pb2_grpc.PrivateDataTableServicer):
    pb2 = private_data_table_pb2
    pb2_grpc = private_data_table_pb2_grpc

    def add(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.add(params)
        return self.dict_to_message(response)

    def transform(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.transform(params)
        return self.dict_to_message(response)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.update(params)
        return self.dict_to_message(response)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        pri_data_table_svc.delete(params)
        return self.empty()

    def load(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.load(params)
        return self.dict_to_message(response)

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.get(params)
        return self.dict_to_message(response)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)
        pri_data_table_svc = PrivateDataTableService(metadata)
        response: dict = pri_data_table_svc.list(params)
        return self.dict_to_message(response)



import logging
from typing import List, Union, Tuple
import pandas as pd

from spaceone.dashboard.error.data_table import (
    ERROR_INVALID_PARAMETER,
    ERROR_NOT_SUPPORTED_OPERATOR,
    ERROR_REQUIRED_PARAMETER,
)
from spaceone.dashboard.manager.data_table_manager import DataTableManager, GRANULARITY
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.model.public_data_table.database import PublicDataTable
from spaceone.dashboard.model.private_data_table.database import PrivateDataTable

_LOGGER = logging.getLogger(__name__)


class DataTransformationManager(DataTableManager):
    def __init__(
        self,
        data_table_type: str,
        operator: str,
        options: dict,
        widget_id: str,
        domain_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if operator not in ["JOIN", "CONCAT", "AGGREGATE", "QUERY", "EVAL"]:
            raise ERROR_NOT_SUPPORTED_OPERATOR(operator=operator)

        self.data_table_type = data_table_type
        self.operator = operator
        self.options = options
        self.widget_id = widget_id
        self.domain_id = domain_id
        self.data_table_vos = self._get_data_table_from_options(operator, options)
        self.data_keys = []
        self.label_keys = []

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        data_info = {}
        labels_info = {}
        for data_table_vo in self.data_table_vos:
            for key, value in data_table_vo.data_info.items():
                if key not in data_info and key in self.data_keys:
                    data_info[key] = value

            for key, value in data_table_vo.labels_info.items():
                if key not in labels_info and key in self.label_keys:
                    labels_info[key] = value

        for key in self.data_keys:
            if key not in data_info:
                data_info[key] = {}

        for key in self.label_keys:
            if key not in labels_info:
                labels_info[key] = {}

        return data_info, labels_info

    def load(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        if self.operator == "JOIN":
            self.join_data_tables(granularity, start, end, vars)
        elif self.operator == "CONCAT":
            self.concat_data_tables(granularity, start, end, vars)
        elif self.operator == "AGGREGATE":
            self.aggregate_data_table(granularity, start, end, vars)
        elif self.operator == "QUERY":
            self.query_data_table(granularity, start, end, vars)
        elif self.operator == "EVAL":
            self.evaluate_data_table(granularity, start, end, vars)

        return self.df

    def join_data_tables(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        how = self.options.get("how", "left")
        if how not in ["left", "right", "inner", "outer"]:
            raise ERROR_INVALID_PARAMETER(
                key="options.JOIN.how", reason="Invalid join type"
            )

        origin_vo = self.data_table_vos[0]
        other_vo = self.data_table_vos[1]
        origin_data_keys = set(origin_vo.data_info.keys())
        other_data_keys = set(other_vo.data_info.keys())

        duplicate_keys = list(origin_data_keys & other_data_keys)
        if len(duplicate_keys) > 0:
            raise ERROR_INVALID_PARAMETER(
                key="options.JOIN.data_tables",
                reason=f"Duplicate data keys: {', '.join(duplicate_keys)}",
            )

        self.data_keys = list(origin_data_keys | other_data_keys)
        origin_label_keys = set(origin_vo.labels_info.keys())
        other_label_keys = set(other_vo.labels_info.keys())
        self.label_keys = list(origin_label_keys | other_label_keys)

        on = list(origin_label_keys & other_label_keys)
        fill_na = {}
        for key in self.data_keys:
            fill_na[key] = 0

        for key in self.label_keys:
            fill_na[key] = ""

        origin_df = self._get_data_table(origin_vo, granularity, start, end, vars)
        other_df = self._get_data_table(other_vo, granularity, start, end, vars)

        if len(other_df) == 0:
            if how in ["left", "outer"]:
                self.df = origin_df
            else:
                self.df = other_df
            return
        elif len(origin_df) == 0:
            if how in ["right", "outer"]:
                self.df = other_df
            else:
                self.df = origin_df
            return

        merged_df = origin_df.merge(other_df, left_on=on, right_on=on, how=how)
        merged_df = merged_df.fillna(value=fill_na)
        self.df = merged_df

    def concat_data_tables(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        origin_vo = self.data_table_vos[0]
        other_vo = self.data_table_vos[1]
        origin_data_keys = set(origin_vo.data_info.keys())
        other_data_keys = set(other_vo.data_info.keys())

        self.data_keys = list(origin_data_keys | other_data_keys)
        origin_label_keys = set(origin_vo.labels_info.keys())
        other_label_keys = set(other_vo.labels_info.keys())
        self.label_keys = list(origin_label_keys | other_label_keys)

        fill_na = {}
        for key in self.data_keys:
            fill_na[key] = 0

        for key in self.label_keys:
            fill_na[key] = ""

        origin_df = self._get_data_table(origin_vo, granularity, start, end, vars)
        other_df = self._get_data_table(other_vo, granularity, start, end, vars)

        merged_df = pd.concat([origin_df, other_df])
        merged_df = merged_df.fillna(value=fill_na)
        self.df = merged_df

    def aggregate_data_table(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        function = dict(self.options.get("function", {}))
        if function == {}:
            raise ERROR_REQUIRED_PARAMETER(key="options.AGGREGATE.function")

        for key, operator in function.items():
            if operator not in ["sum", "mean", "max", "min", "count"]:
                raise ERROR_INVALID_PARAMETER(
                    key="options.AGGREGATE.function",
                    reason=f"Invalid function type: {operator}",
                )

        group_by = list(self.options.get("group_by", []))

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(function.keys())
        self.label_keys = group_by

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        if len(group_by) > 0:
            self.df = df.groupby(group_by).agg(function).reset_index()
        else:
            self.df = df.agg(function).to_frame().T

    def query_data_table(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        conditions = self.options.get("conditions", [])

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(origin_vo.data_info.keys())
        self.label_keys = list(origin_vo.labels_info.keys())

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        for condition in conditions:
            try:
                df = df.query(condition)
            except Exception as e:
                _LOGGER.error(f"[query_data_table] query error: {e}")
                raise ERROR_INVALID_PARAMETER(
                    key="options.QUERY.conditions", reason=condition
                )

        self.df = df

    def evaluate_data_table(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> None:
        expressions = self.options.get("expressions", [])

        origin_vo = self.data_table_vos[0]
        self.data_keys = list(origin_vo.data_info.keys())
        self.label_keys = list(origin_vo.labels_info.keys())

        df = self._get_data_table(origin_vo, granularity, start, end, vars)

        for expression in expressions:
            try:
                key, value = expression.split("=")
                self.data_keys = list(set(self.data_keys) | {key})

                df = df.eval(expression)

            except Exception as e:
                _LOGGER.error(f"[evaluate_data_table] eval error: {e}")
                raise ERROR_INVALID_PARAMETER(
                    key="options.EVAL.expressions", reason=expression
                )

        self.df = df

    def _get_data_table(
        self,
        data_table_vo: Union[PublicDataTable, PrivateDataTable],
        granularity: GRANULARITY,
        start: str,
        end: str,
        vars: dict,
    ) -> pd.DataFrame:
        if data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                self.data_table_type,
                data_table_vo.source_type,
                data_table_vo.options,
                data_table_vo.widget_id,
                data_table_vo.domain_id,
            )
            return ds_mgr.load(granularity, start, end, vars)
        else:
            operator = data_table_vo.operator
            options = data_table_vo.options.get(operator, {})
            ds_mgr = DataTransformationManager(
                self.data_table_type,
                data_table_vo.operator,
                options,
                data_table_vo.widget_id,
                data_table_vo.domain_id,
            )
            return ds_mgr.load(granularity, start, end, vars)

    def _get_data_table_from_options(
        self,
        operator: str,
        options: dict,
    ) -> List[Union[PublicDataTable, PrivateDataTable]]:
        parent_dt_vos = []

        if self.data_table_type == "PUBLIC":
            data_table_mgr = PublicDataTableManager()
        else:
            data_table_mgr = PrivateDataTableManager()

        if operator in ["JOIN", "CONCAT"]:
            if data_tables := options.get("data_tables"):
                if len(data_tables) != 2:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.data_tables",
                        reason="It should have 2 data tables.",
                    )

                for data_table_id in data_tables:
                    if self.data_table_type == "PUBLIC":
                        parent_dt_vo = data_table_mgr.get_public_data_table(
                            data_table_id,
                            self.domain_id,
                        )
                    else:
                        parent_dt_vo = data_table_mgr.get_private_data_table(
                            data_table_id,
                            self.domain_id,
                        )

                    if parent_dt_vo.widget_id != self.widget_id:
                        raise ERROR_INVALID_PARAMETER(
                            key="options.data_tables",
                            reason="It should have same widget_id.",
                        )
                    parent_dt_vos.append(parent_dt_vo)

            else:
                raise ERROR_REQUIRED_PARAMETER(key=f"options.{operator}.data_tables")
        else:
            if data_table_id := options.get("data_table_id"):
                if self.data_table_type == "PUBLIC":
                    parent_dt_vo = data_table_mgr.get_public_data_table(
                        data_table_id,
                        self.domain_id,
                    )
                else:
                    parent_dt_vo = data_table_mgr.get_private_data_table(
                        data_table_id,
                        self.domain_id,
                    )

                if parent_dt_vo.widget_id != self.widget_id:
                    raise ERROR_INVALID_PARAMETER(
                        key="options.data_table_id",
                        reason="It should have same widget_id.",
                    )
                parent_dt_vos.append(parent_dt_vo)
            else:
                raise ERROR_REQUIRED_PARAMETER(key="options.data_table_id")

        return parent_dt_vos



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.manager.private_widget_manager import PrivateWidgetManager
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.data_table_manager.data_transformation_manager import (
    DataTransformationManager,
)
from spaceone.dashboard.model.private_data_table.request import *
from spaceone.dashboard.model.private_data_table.response import *
from spaceone.dashboard.model.private_data_table.database import PrivateDataTable

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateDataTableService(BaseService):
    resource = "PrivateDataTable"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_data_table_mgr = PrivateDataTableManager()

    @transaction(
        permission="dashboard:PrivateDataTable.write",
        role_types=["USER"],
    )
    @convert_model
    def add(
        self, params: PrivateDataTableAddRequest
    ) -> Union[PrivateDataTableResponse, dict]:
        """Add private data table

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'source_type': 'str',           # required
                'options': 'dict',              # required
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str',             # injected from auth (required)
            }

        Returns:
            PrivateDataTableResponse:
        """

        pri_widget_mgr = PrivateWidgetManager()
        pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        ds_mgr = DataSourceManager(
            "PRIVATE",
            params.source_type,
            params.options,
            params.widget_id,
            params.domain_id,
        )

        # Load data source to verify options
        ds_mgr.load()

        # Get data and labels info from options
        data_info, labels_info = ds_mgr.get_data_and_labels_info()

        params_dict = params.dict()
        params_dict["data_type"] = "ADDED"
        params_dict["data_info"] = data_info
        params_dict["labels_info"] = labels_info

        pri_data_table_vo = self.pri_data_table_mgr.create_private_data_table(
            params_dict
        )

        return PrivateDataTableResponse(**pri_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDataTable.write",
        role_types=["USER"],
    )
    @convert_model
    def transform(
        self, params: PrivateDataTableTransformRequest
    ) -> Union[PrivateDataTableResponse, dict]:
        """Add private data table

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'operator': 'str',              # required
                'options': 'dict',              # required
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str',             # injected from auth (required)
            }

        Returns:
            PrivateDataTableResponse:
        """

        pri_widget_mgr = PrivateWidgetManager()
        pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        operator = params.operator
        options = params.options.get(operator, {})
        dt_mgr = DataTransformationManager(
            "PRIVATE", operator, options, params.widget_id, params.domain_id
        )

        # Load data table to verify options
        dt_mgr.load()

        # Get data and labels info from options
        data_info, labels_info = dt_mgr.get_data_and_labels_info()

        params_dict = params.dict()
        params_dict["data_type"] = "TRANSFORMED"
        params_dict["data_info"] = data_info
        params_dict["labels_info"] = labels_info

        pri_data_table_vo = self.pri_data_table_mgr.create_private_data_table(
            params_dict
        )

        return PrivateDataTableResponse(**pri_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDataTable.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateDataTableUpdateRequest
    ) -> Union[PrivateDataTableResponse, dict]:
        """Update private data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'name': 'str',
                'options': 'dict',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDataTableResponse:
        """

        pri_data_table_vo: PrivateDataTable = (
            self.pri_data_table_mgr.get_private_data_table(
                params.data_table_id, params.domain_id, params.user_id
            )
        )

        params_dict = params.dict(exclude_unset=True)

        if options := params_dict.get("options"):
            if pri_data_table_vo.data_type == "ADDED":
                ds_mgr = DataSourceManager(
                    "PRIVATE",
                    pri_data_table_vo.source_type,
                    options,
                    pri_data_table_vo.widget_id,
                    pri_data_table_vo.domain_id,
                )

                # Load data source to verify options
                ds_mgr.load()

                # Get data and labels info from options
                data_info, labels_info = ds_mgr.get_data_and_labels_info()
                params_dict["data_info"] = data_info
                params_dict["labels_info"] = labels_info

                # change timediff format
                if timediff := options.get("timediff"):
                    if years := timediff.get("years"):
                        options["timediff"] = {"years": years}
                    elif months := timediff.get("months"):
                        options["timediff"] = {"months": months}
                    elif days := timediff.get("days"):
                        options["timediff"] = {"days": days}

                    params_dict["options"] = options
            else:
                operator = pri_data_table_vo.operator
                operator_options = options.get(operator, {})
                dt_mgr = DataTransformationManager(
                    "PRIVATE",
                    operator,
                    operator_options,
                    pri_data_table_vo.widget_id,
                    pri_data_table_vo.domain_id,
                )

                # Load data table to verify options
                dt_mgr.load()

                # Get data and labels info from options
                data_info, labels_info = dt_mgr.get_data_and_labels_info()

                params_dict = params.dict()
                params_dict["data_type"] = "TRANSFORMED"
                params_dict["data_info"] = data_info
                params_dict["labels_info"] = labels_info
                params_dict["options"] = {
                    operator: operator_options,
                }

        pri_data_table_vo = self.pri_data_table_mgr.update_private_data_table_by_vo(
            params_dict, pri_data_table_vo
        )

        return PrivateDataTableResponse(**pri_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDataTable.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateDataTableDeleteRequest) -> None:
        """Delete private data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_data_table_vo: PrivateDataTable = (
            self.pri_data_table_mgr.get_private_data_table(
                params.data_table_id,
                params.domain_id,
                params.user_id,
            )
        )

        self.pri_data_table_mgr.delete_private_data_table_by_vo(pri_data_table_vo)

    @transaction(
        permission="dashboard:PrivateDataTable.write",
        role_types=["USER"],
    )
    @convert_model
    def load(self, params: PrivateDataTableLoadRequest) -> dict:
        """Load private data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'granularity': 'str',           # required
                'start': 'str',
                'end': 'str',
                'sort': 'list',
                'page': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_data_table_vo: PrivateDataTable = (
            self.pri_data_table_mgr.get_private_data_table(
                params.data_table_id,
                params.domain_id,
                params.user_id,
            )
        )

        if pri_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PRIVATE",
                pri_data_table_vo.source_type,
                pri_data_table_vo.options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            ds_mgr.load(
                params.granularity,
                params.start,
                params.end,
            )
            return ds_mgr.response(params.sort, params.page)

        else:
            operator = pri_data_table_vo.operator
            options = pri_data_table_vo.options.get(operator, {})
            widget_id = pri_data_table_vo.widget_id
            domain_id = pri_data_table_vo.domain_id

            dt_mgr = DataTransformationManager(
                "PUBLIC",
                operator,
                options,
                widget_id,
                domain_id,
            )
            dt_mgr.load(params.granularity, params.start, params.end)
            return dt_mgr.response(params.sort, params.page)

    @transaction(
        permission="dashboard:PrivateDataTable.read",
        role_types=["USER"],
    )
    @convert_model
    def get(
        self, params: PrivateDataTableGetRequest
    ) -> Union[PrivateDataTableResponse, dict]:
        """Get private data table

        Args:
            params (dict): {
                'data_table_id': 'str',         # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateDataTableResponse:
        """

        pri_data_table_vo: PrivateDataTable = (
            self.pri_data_table_mgr.get_private_data_table(
                params.data_table_id,
                params.domain_id,
                params.user_id,
            )
        )

        return PrivateDataTableResponse(**pri_data_table_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateDataTable.read",
        role_types=["USER"],
    )
    @append_query_filter(
        [
            "widget_id",
            "data_table_id",
            "name",
            "data_type",
            "source_type",
            "operator",
            "domain_id",
            "user_id",
        ]
    )
    @append_keyword_filter(["data_table_id", "name"])
    @convert_model
    def list(
        self, params: PrivateDataTableSearchQueryRequest
    ) -> Union[PrivateDataTablesResponse, dict]:
        """List private data tables

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'widget_id': 'str',                             # required
                'data_table_id': 'str',
                'name': 'str',
                'data_type': 'str',
                'source_type': 'str',
                'operator': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateDataTablesResponse:
        """

        query = params.query or {}
        (
            pri_data_table_vos,
            total_count,
        ) = self.pri_data_table_mgr.list_private_data_tables(query)
        pri_data_tables_info = [
            pri_data_table_vo.to_dict() for pri_data_table_vo in pri_data_table_vos
        ]
        return PrivateDataTablesResponse(
            results=pri_data_tables_info, total_count=total_count
        )



import logging
from typing import Union, Literal, Tuple
import pandas as pd

from spaceone.core.manager import BaseManager
from spaceone.dashboard.error.data_table import (
    ERROR_REQUIRED_PARAMETER,
)
from spaceone.dashboard.error.data_table import (
    ERROR_QUERY_OPTION,
    ERROR_NOT_SUPPORTED_QUERY_OPTION,
)

_LOGGER = logging.getLogger(__name__)
GRANULARITY = Literal["DAILY", "MONTHLY", "YEARLY"]


class DataTableManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.df: Union[pd.DataFrame, None] = None

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        raise NotImplementedError()

    def load(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def load_from_widget(self, query: dict, vars: dict = None) -> dict:
        self._check_query(query)
        granularity = query["granularity"]
        start = query["start"]
        end = query["end"]
        group_by = query.get("group_by")
        filter = query.get("filter")
        fields = query.get("fields")
        field_group = query.get("field_group")
        sort = query.get("sort")
        page = query.get("page")

        self.load(
            granularity,
            start,
            end,
            vars=vars,
        )

        if filter:
            self.apply_filter(filter)

        self.apply_group_by(fields, group_by)

        if field_group:
            self.apply_field_group(field_group, fields)

            if sort:
                changed_sort = []
                for condition in sort:
                    key = condition.get("key")
                    desc = condition.get("desc", False)

                    if key in fields:
                        changed_sort.append({"key": f"_total_{key}", "desc": desc})
                    else:
                        changed_sort.append(condition)

                sort = changed_sort

        return self.response(sort, page)

    @staticmethod
    def _check_query(query: dict) -> None:
        if "granularity" not in query:
            raise ERROR_REQUIRED_PARAMETER(key="query.granularity")

        if "start" not in query:
            raise ERROR_REQUIRED_PARAMETER(key="query.start")

        if "end" not in query:
            raise ERROR_REQUIRED_PARAMETER(key="query.end")

        if "fields" not in query:
            raise ERROR_REQUIRED_PARAMETER(key="query.fields")

        if "select" in query:
            raise ERROR_NOT_SUPPORTED_QUERY_OPTION(key="query.select")

        if "filter_or" in query:
            raise ERROR_NOT_SUPPORTED_QUERY_OPTION(key="query.filter_or")

    def response(self, sort: list = None, page: dict = None) -> dict:
        total_count = len(self.df)

        if sort:
            self.apply_sort(sort)

        if page:
            self.apply_page(page)

        return {
            "results": self.df.to_dict(orient="records"),
            "total_count": total_count,
        }

    def apply_filter(self, filter: list) -> None:
        if len(self.df) > 0:
            for condition in filter:
                key = condition.get("key", condition.get("k"))
                operator = condition.get("operator", condition.get("o"))
                value = condition.get("value", condition.get("v"))

                if operator in ["in", "not in"]:
                    if not isinstance(value, list):
                        raise ERROR_QUERY_OPTION(key="filter")

                if key and operator and value:
                    try:
                        if operator == "in":
                            self.df = self.df[self.df[key].isin(value)]
                        elif operator == "not_in":
                            self.df = self.df[~self.df[key].isin(value)]
                        elif operator == "eq":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} == {value}")
                            else:
                                self.df = self.df.query(f"{key} == '{value}'")
                        elif operator == "not":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} != {value}")
                            else:
                                self.df = self.df.query(f"{key} != '{value}'")
                        elif operator == "gt":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} > {value}")
                            else:
                                self.df = self.df.query(f"{key} > '{value}'")
                        elif operator == "gte":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} >= {value}")
                            else:
                                self.df = self.df.query(f"{key} >= '{value}'")
                        elif operator == "lt":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} < {value}")
                            else:
                                self.df = self.df.query(f"{key} < '{value}'")
                        elif operator == "lte":
                            if isinstance(value, int) or isinstance(value, float):
                                self.df = self.df.query(f"{key} <= {value}")
                            else:
                                self.df = self.df.query(f"{key} <= '{value}'")
                        elif operator == "contain":
                            self.df = self.df[self.df[key].str.contains(str(value))]
                        elif operator == "not_contain":
                            self.df = self.df[~self.df[key].str.contains(str(value))]
                        else:
                            raise ERROR_NOT_SUPPORTED_QUERY_OPTION(
                                key=f"filter.operator.{operator}"
                            )
                    except Exception as e:
                        raise ERROR_QUERY_OPTION(key="filter")
                else:
                    raise ERROR_QUERY_OPTION(key="filter")

    def apply_group_by(self, fields: dict, group_by: list = None) -> None:
        if len(self.df) > 0:
            columns = list(fields.keys())
            if group_by:
                columns.extend(group_by)

            self.df = self.df[columns]

            agg_options = {}
            for name, options in fields.items():
                operator = options.get("operator", "sum")
                if operator not in ["sum", "average", "max", "min"]:
                    raise ERROR_NOT_SUPPORTED_QUERY_OPTION(
                        key=f"fields.operator.{operator}"
                    )

                if operator == "average":
                    operator = "mean"

                agg_options[name] = operator

            if group_by:
                self.df = self.df.groupby(group_by).agg(agg_options).reset_index()
            else:
                self.df = self.df.agg(agg_options).to_frame().T

    def apply_field_group(self, field_group: list, fields: dict) -> None:
        if len(self.df) > 0:
            data_fields = list(fields.keys())
            agg_fields = set(data_fields + field_group)
            group_by = list(set(self.df.columns) - agg_fields)
            agg_options = {}
            for field in agg_fields:
                agg_options[field] = lambda x: list(x)

            self.df = self.df.groupby(group_by).agg(agg_options).reset_index()
            changed_data = []
            for row in self.df.to_dict(orient="records"):
                changed_row = {}
                for data_field in data_fields:
                    changed_row[data_field] = []

                for key, value in row.items():
                    if key in group_by:
                        changed_row[key] = value
                    elif key in fields:
                        operator = fields[key].get("operator", "sum")
                        if operator == "sum":
                            changed_row[f"_total_{key}"] = sum(value)
                        elif operator == "average":
                            changed_row[f"_total_{key}"] = sum(value) / len(value)
                        elif operator == "max":
                            changed_row[f"_total_{key}"] = max(value)
                        elif operator == "min":
                            changed_row[f"_total_{key}"] = min(value)

                        for idx, v in enumerate(value):
                            data = {"value": v}
                            for fg in field_group:
                                data[fg] = row[fg][idx]

                            changed_row[key].append(data)

                changed_data.append(changed_row)

            self.df = pd.DataFrame(changed_data)

    def apply_sort(self, sort: list) -> None:
        if len(self.df) > 0:
            keys = []
            ascendings = []

            for sort_option in sort:
                key = sort_option.get("key")
                ascending = not sort_option.get("desc", False)

                if key:
                    keys.append(key)
                    ascendings.append(ascending)

            try:
                self.df = self.df.sort_values(by=keys, ascending=ascendings)
            except Exception as e:
                _LOGGER.error(f"[_sort] Sort Error: {e}")
                raise ERROR_QUERY_OPTION(key="sort")

    def apply_page(self, page: dict) -> None:
        if len(self.df) > 0:
            if limit := page.get("limit"):
                if limit > 0:
                    start = page.get("start", 1)
                    if start < 1:
                        start = 1

                    self.df = self.df.iloc[start - 1 : start + limit - 1]



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.dashboard.manager.private_widget_manager import PrivateWidgetManager
from spaceone.dashboard.manager.private_dashboard_manager import PrivateDashboardManager
from spaceone.dashboard.manager.private_data_table_manager import (
    PrivateDataTableManager,
)
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.data_table_manager.data_transformation_manager import (
    DataTransformationManager,
)
from spaceone.dashboard.model.private_widget.request import *
from spaceone.dashboard.model.private_widget.response import *
from spaceone.dashboard.error.dashboard import (
    ERROR_NOT_SUPPORTED_VERSION,
    ERROR_INVALID_PARAMETER,
)

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PrivateWidgetService(BaseService):
    resource = "PrivateWidget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pri_widget_mgr = PrivateWidgetManager()

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def create(
        self, params: PrivateWidgetCreateRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Create private widget

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'name': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str',             # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        pri_dashboard_mgr = PrivateDashboardManager()
        pri_dashboard_vo = pri_dashboard_mgr.get_private_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.user_id,
        )

        if pri_dashboard_vo.version == "1.0":
            raise ERROR_NOT_SUPPORTED_VERSION(version=pri_dashboard_vo.version)

        pri_widget_vo = self.pri_widget_mgr.create_private_widget(params.dict())

        return PrivateWidgetResponse(**pri_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def update(
        self, params: PrivateWidgetUpdateRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Update private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'data_table_id': 'str',
                'tags': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        if params.data_table_id is not None:
            pri_data_table_mgr = PrivateDataTableManager()
            pri_data_table_vo = pri_data_table_mgr.get_private_data_table(
                params.data_table_id,
                params.domain_id,
                params.user_id,
            )

            if pri_data_table_vo.widget_id != params.widget_id:
                raise ERROR_INVALID_PARAMETER(
                    key="data_table_id",
                    reason="Data table is not belong to this widget.",
                )

        pri_widget_vo = self.pri_widget_mgr.update_private_widget_by_vo(
            params.dict(exclude_unset=True), pri_widget_vo
        )

        return PrivateWidgetResponse(**pri_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def delete(self, params: PrivateWidgetDeleteRequest) -> None:
        """Delete private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        self.pri_widget_mgr.delete_private_widget_by_vo(pri_widget_vo)

    @transaction(
        permission="dashboard:PrivateWidget.write",
        role_types=["USER"],
    )
    @convert_model
    def load(self, params: PrivateWidgetLoadRequest) -> dict:
        """Load private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'query': 'dict (spaceone.api.core.v1.AnalyzeQuery)', # required
                'vars': 'dict',
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            None
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        if pri_widget_vo.data_table_id is None:
            raise ERROR_INVALID_PARAMETER(
                key="widget_id", reason="Data table is not set."
            )

        pri_data_table_mgr = PrivateDataTableManager()
        pri_data_table_vo = pri_data_table_mgr.get_private_data_table(
            pri_widget_vo.data_table_id,
            params.domain_id,
            params.user_id,
        )

        if pri_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PRIVATE",
                pri_data_table_vo.source_type,
                pri_data_table_vo.options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return ds_mgr.load_from_widget(
                params.query,
                params.vars,
            )
        else:
            operator = pri_data_table_vo.operator
            options = pri_data_table_vo.options.get(operator, {})

            dt_mgr = DataTransformationManager(
                "PRIVATE",
                pri_data_table_vo.operator,
                options,
                pri_data_table_vo.widget_id,
                pri_data_table_vo.domain_id,
            )
            return dt_mgr.load_from_widget(
                params.query,
                params.vars,
            )

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @convert_model
    def get(
        self, params: PrivateWidgetGetRequest
    ) -> Union[PrivateWidgetResponse, dict]:
        """Get private widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'user_id': 'str',               # injected from auth (required)
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PrivateWidgetResponse:
        """

        pri_widget_vo = self.pri_widget_mgr.get_private_widget(
            params.widget_id,
            params.domain_id,
            params.user_id,
        )

        return PrivateWidgetResponse(**pri_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PrivateWidget.read",
        role_types=["USER"],
    )
    @append_query_filter(
        [
            "dashboard_id",
            "widget_id",
            "name",
            "domain_id",
            "user_id",
        ]
    )
    @append_keyword_filter(["widget_id", "name"])
    @convert_model
    def list(
        self, params: PrivateWidgetSearchQueryRequest
    ) -> Union[PrivateWidgetsResponse, dict]:
        """List private widgets

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',                          # required
                'widget_id': 'str',
                'name': 'str',
                'user_id': 'str',                               # injected from auth (required)
                'domain_id': 'str',                             # injected from auth (required)
            }

        Returns:
            PrivateWidgetsResponse:
        """

        query = params.query or {}
        pri_widget_vos, total_count = self.pri_widget_mgr.list_private_widgets(query)
        pri_widgets_info = [pri_widget_vo.to_dict() for pri_widget_vo in pri_widget_vos]
        return PrivateWidgetsResponse(results=pri_widgets_info, total_count=total_count)



import logging
import copy
from typing import Tuple
from datetime import datetime
from dateutil.relativedelta import relativedelta
import pandas as pd

from spaceone.dashboard.manager.data_table_manager import DataTableManager, GRANULARITY
from spaceone.dashboard.manager.cost_analysis_manager import CostAnalysisManager
from spaceone.dashboard.manager.inventory_manager import InventoryManager
from spaceone.dashboard.error.data_table import *

_LOGGER = logging.getLogger(__name__)


class DataSourceManager(DataTableManager):
    def __init__(
        self,
        data_table_type: str,
        source_type: str,
        options: dict,
        widget_id: str,
        domain_id: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if source_type not in ["COST", "ASSET"]:
            raise ERROR_NOT_SUPPORTED_SOURCE_TYPE(source_type=source_type)

        self.data_table_type = data_table_type
        self.cost_analysis_mgr = CostAnalysisManager()
        self.inventory_mgr = InventoryManager()
        self.source_type = source_type
        self.options = options
        self.widget_id = widget_id
        self.domain_id = domain_id

        self.data_name = options.get("data_name")
        if self.data_name is None:
            raise ERROR_REQUIRED_PARAMETER(key="options.data_name")

        self.data_unit = options.get("data_unit")
        self.date_format = options.get("date_format", "SINGLE")
        self.timediff = options.get("timediff")
        self.group_by = options.get("group_by")
        self.filter = options.get("filter")
        self.filter_or = options.get("filter_or")
        self.additional_labels = options.get("additional_labels")

    def get_data_and_labels_info(self) -> Tuple[dict, dict]:
        data_info = {self.data_name: {}}

        if self.data_unit:
            data_info[self.data_name]["unit"] = self.data_unit

        labels_info = {}

        if self.group_by:
            for group_option in copy.deepcopy(self.group_by):
                if isinstance(group_option, dict):
                    group_name = group_option.get("name")
                    group_key = group_option.get("key")
                    if "." in group_key:
                        group_key = group_key.split(".")[-1]

                    name = group_name or group_key
                    if name is None:
                        raise ERROR_REQUIRED_PARAMETER(key="options.group_by.key")

                    if group_name:
                        del group_option["name"]

                    if group_key:
                        del group_option["key"]

                    labels_info[group_name] = group_option
                else:
                    labels_info[group_option] = {}

        if self.additional_labels:
            for key in self.additional_labels.keys():
                labels_info[key] = {}

        if self.date_format == "SINGLE":
            labels_info["Date"] = {}
        else:
            labels_info["Year"] = {}
            labels_info["Month"] = {}
            labels_info["Day"] = {}

        return data_info, labels_info

    def load(
        self,
        granularity: GRANULARITY = "DAILY",
        start: str = None,
        end: str = None,
        vars: dict = None,
    ) -> pd.DataFrame:
        start, end = self._get_time_from_granularity(granularity, start, end)

        if self.timediff:
            start, end = self._change_time(start, end)

        if self.source_type == "COST":
            self._analyze_cost(granularity, start, end, vars)
        elif self.source_type == "ASSET":
            self._analyze_asset(granularity, start, end, vars)

        if additional_labels := self.options.get("additional_labels"):
            self._add_labels(additional_labels)

        return self.df

    def _add_labels(self, labels: dict) -> None:
        for key, value in labels.items():
            self.df[key] = value

    def _analyze_asset(
        self,
        granularity: GRANULARITY,
        start: str,
        end: str,
        vars: dict = None,
    ) -> None:
        asset_info = self.options.get("ASSET", {})
        metric_id = asset_info.get("metric_id")
        data_key = "value"

        if metric_id is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.ASSET.metric_id")

        query = self._make_query(
            data_key,
            granularity,
            start,
            end,
            vars=vars,
        )

        params = {"metric_id": metric_id, "query": query}

        response = self.inventory_mgr.analyze_metric_data(params)
        results = response.get("results", [])

        results = self._change_datetime_format(results)

        self.df = pd.DataFrame(results)

    def _analyze_cost(
        self,
        granularity: GRANULARITY,
        start: str,
        end: str,
        vars: dict = None,
    ) -> None:
        cost_info = self.options.get("COST", {})
        data_source_id = cost_info.get("data_source_id")
        data_key = cost_info.get("data_key")

        if data_source_id is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.COST.data_source_id")

        if data_key is None:
            raise ERROR_REQUIRED_PARAMETER(parameter="options.COST.data_key")

        query = self._make_query(
            data_key,
            granularity,
            start,
            end,
            vars=vars,
        )

        params = {"data_source_id": data_source_id, "query": query}

        response = self.cost_analysis_mgr.analyze_cost(params)
        results = response.get("results", [])

        results = self._change_datetime_format(results)

        self.df = pd.DataFrame(results)

    def _change_datetime_format(self, results: list) -> list:
        changed_results = []
        for result in results:
            if date := result.get("date"):
                if self.date_format == "SINGLE":
                    result["Date"] = date
                else:
                    if len(date) == 4:
                        result["Year"] = date
                    elif len(date) == 7:
                        year, month = date.split("-")
                        result["Year"] = year
                        result["Month"] = month
                    elif len(date) == 10:
                        year, month, day = date.split("-")
                        result["Year"] = year
                        result["Month"] = month
                        result["Day"] = day

                del result["date"]
            changed_results.append(result)
        return changed_results

    def _change_time(self, start: str, end: str) -> Tuple[str, str]:
        start_len = len(start)
        end_len = len(end)
        start_time = self._get_datetime_from_str(start)
        end_time = self._get_datetime_from_str(end)

        years = self.timediff.get("years", 0)
        months = self.timediff.get("months", 0)
        days = self.timediff.get("days", 0)

        if years:
            start_time = start_time + relativedelta(years=years)
            end_time = end_time + relativedelta(years=years)
        elif months:
            start_time = start_time + relativedelta(months=months)
            end_time = end_time + relativedelta(months=months)
        elif days:
            start_time = start_time + relativedelta(days=days)
            end_time = end_time + relativedelta(days=days)

        return (
            self._change_str_from_datetime(start_time, start_len),
            self._change_str_from_datetime(end_time, end_len),
        )

    @staticmethod
    def _change_str_from_datetime(dt: datetime, date_str_len: int) -> str:
        if date_str_len == 4:
            return dt.strftime("%Y")
        elif date_str_len == 7:
            return dt.strftime("%Y-%m")
        else:
            return dt.strftime("%Y-%m-%d")

    @staticmethod
    def _get_datetime_from_str(datetime_str: str, is_end: bool = False) -> datetime:
        if len(datetime_str) == 4:
            dt = datetime.strptime(datetime_str, "%Y")
            if is_end:
                dt = dt + relativedelta(years=1) - relativedelta(days=1)

        elif len(datetime_str) == 7:
            dt = datetime.strptime(datetime_str, "%Y-%m")
            if is_end:
                dt = dt + relativedelta(months=1) - relativedelta(days=1)
        else:
            dt = datetime.strptime(datetime_str, "%Y-%m-%d")

        return dt

    @staticmethod
    def _get_time_from_granularity(
        granularity: GRANULARITY,
        start: str = None,
        end: str = None,
    ) -> Tuple[str, str]:
        if start and end:
            return start, end
        else:
            now = datetime.utcnow()

            if granularity == "YEARLY":
                end_time = now.replace(month=1, day=1, hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(years=2)
                return start_time.strftime("%Y"), end_time.strftime("%Y")

            elif granularity == "MONTHLY":
                end_time = now.replace(day=1, hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(months=5)
                return start_time.strftime("%Y-%m"), end_time.strftime("%Y-%m")

            else:
                end_time = now.replace(hour=0, minute=0, second=0)
                start_time = end_time - relativedelta(days=29)
                return start_time.strftime("%Y-%m-%d"), end_time.strftime("%Y-%m-%d")

    def _make_query(
        self,
        data_key: str,
        granularity: GRANULARITY,
        start: str,
        end: str,
        vars: dict = None,
    ):
        if vars:
            self.filter = self.filter or []
            for key, value in vars.items():
                if key in ["workspace_id", "project_id", "service_account_id"]:
                    if isinstance(value, list):
                        self.filter.append(
                            {"key": key, "value": value, "operator": "in"}
                        )
                    else:
                        self.filter.append(
                            {"key": key, "value": value, "operator": "eq"}
                        )

        return {
            "granularity": granularity,
            "start": start,
            "end": end,
            "group_by": self.group_by,
            "filter": self.filter,
            "filter_or": self.filter_or,
            "fields": {self.data_name: {"key": data_key, "operator": "sum"}},
        }



from datetime import datetime
from typing import Union, List
from pydantic import BaseModel
from spaceone.core import utils

__all__ = ["PrivateWidgetResponse", "PrivateWidgetsResponse"]


class PrivateWidgetResponse(BaseModel):
    widget_id: Union[str, None] = None
    name: Union[str, None] = None
    description: Union[str, None] = None
    widget_type: Union[str, None] = None
    size: Union[str, None] = None
    options: Union[dict, None] = None
    tags: Union[dict, None] = None
    data_table_id: Union[str, None] = None
    dashboard_id: Union[str, None] = None
    user_id: Union[str, None] = None
    domain_id: Union[str, None] = None
    created_at: Union[datetime, None] = None
    updated_at: Union[datetime, None] = None

    def dict(self, *args, **kwargs):
        data = super().dict(*args, **kwargs)
        data["created_at"] = utils.datetime_to_iso8601(data["created_at"])
        data["updated_at"] = utils.datetime_to_iso8601(data["updated_at"])
        return data


class PrivateWidgetsResponse(BaseModel):
    results: List[PrivateWidgetResponse]
    total_count: int



from typing import Union, Literal
from pydantic import BaseModel

__all__ = [
    "PublicDataTableAddRequest",
    "PublicDataTableTransformRequest",
    "PublicDataTableUpdateRequest",
    "PublicDataTableDeleteRequest",
    "PublicDataTableLoadRequest",
    "PublicDataTableGetRequest",
    "PublicDataTableSearchQueryRequest",
]

Granularity = Literal["DAILY", "MONTHLY", "YEARLY"]


class PublicDataTableAddRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    source_type: str
    options: dict
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableTransformRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    operator: str
    options: dict
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableUpdateRequest(BaseModel):
    data_table_id: str
    name: Union[str, None] = None
    options: Union[dict, None] = None
    tags: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableDeleteRequest(BaseModel):
    data_table_id: str
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableLoadRequest(BaseModel):
    data_table_id: str
    granularity: Granularity
    start: Union[str, None] = None
    end: Union[str, None] = None
    sort: Union[list, None] = None
    page: Union[dict, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableGetRequest(BaseModel):
    data_table_id: str
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    widget_id: str
    data_table_id: Union[str, None] = None
    name: Union[str, None] = None
    data_type: Union[str, None] = None
    source_type: Union[str, None] = None
    operator: Union[str, None] = None
    project_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDataTableStatQueryRequest(BaseModel):
    query: dict
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None



from typing import Union
from pydantic import BaseModel

__all__ = [
    "PrivateWidgetCreateRequest",
    "PrivateWidgetUpdateRequest",
    "PrivateWidgetDeleteRequest",
    "PrivateWidgetLoadRequest",
    "PrivateWidgetGetRequest",
    "PrivateWidgetSearchQueryRequest",
]


class PrivateWidgetCreateRequest(BaseModel):
    dashboard_id: str
    name: Union[str, None] = None
    description: Union[str, None] = None
    widget_type: Union[str, None] = None
    size: Union[str, None] = None
    options: Union[dict, None] = None
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateWidgetUpdateRequest(BaseModel):
    widget_id: str
    name: Union[str, None] = None
    description: Union[str, None] = None
    widget_type: Union[str, None] = None
    size: Union[str, None] = None
    options: Union[dict, None] = None
    data_table_id: Union[str, None] = None
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateWidgetDeleteRequest(BaseModel):
    widget_id: str
    user_id: str
    domain_id: str


class PrivateWidgetLoadRequest(BaseModel):
    widget_id: str
    query: dict
    vars: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateWidgetGetRequest(BaseModel):
    widget_id: str
    user_id: str
    domain_id: str


class PrivateWidgetSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    dashboard_id: str
    widget_id: Union[str, None] = None
    name: Union[str, None] = None
    user_id: str
    domain_id: str


class PrivateWidgetStatQueryRequest(BaseModel):
    query: dict
    user_id: str
    domain_id: str



from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector


class IdentityManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.identity_conn: SpaceConnector = self.locator.get_connector(
            "SpaceConnector", service="identity"
        )

    def check_workspace(self, workspace_id: str, domain_id: str) -> None:
        system_token = config.get_global("TOKEN")

        return self.identity_conn.dispatch(
            "Workspace.check",
            {"workspace_id": workspace_id, "domain_id": domain_id},
            token=system_token,
        )

    def get_project(self, project_id: str) -> dict:
        return self.identity_conn.dispatch("Project.get", {"project_id": project_id})



from typing import Union, List, Literal, Any
from pydantic import BaseModel

__all__ = [
    "PublicDashboardCreateRequest",
    "PublicDashboardUpdateRequest",
    "PublicDashboardDeleteRequest",
    "PublicDashboardGetRequest",
    "PublicDashboardSearchQueryRequest",
    "PublicDashboardStatQueryRequest",
    "ResourceGroup",
]

ResourceGroup = Literal["DOMAIN", "WORKSPACE", "PROJECT"]


class PublicDashboardCreateRequest(BaseModel):
    name: str
    description: Union[str, None] = None
    layouts: Union[List[Any], None] = None
    vars: Union[dict, None] = None
    options: Union[dict, None] = None
    variables: Union[dict, None] = None
    variables_schema: Union[dict, None] = None
    labels: Union[List[str], None] = None
    tags: Union[dict, None] = None
    folder_id: Union[str, None] = None
    resource_group: ResourceGroup
    project_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDashboardUpdateRequest(BaseModel):
    dashboard_id: str
    name: Union[str, None] = None
    description: Union[str, None] = None
    layouts: Union[List[Any], None] = None
    vars: Union[dict, None] = None
    settings: Union[dict, None] = None
    variables: Union[dict, None] = None
    variables_schema: Union[dict, None] = None
    labels: Union[List[str], None] = None
    tags: Union[dict, None] = None
    folder_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDashboardDeleteRequest(BaseModel):
    dashboard_id: str
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDashboardGetRequest(BaseModel):
    dashboard_id: str
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDashboardSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    dashboard_id: Union[str, None] = None
    name: Union[str, None] = None
    folder_id: Union[str, None] = None
    project_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None


class PublicDashboardStatQueryRequest(BaseModel):
    query: dict
    workspace_id: Union[str, None] = None
    domain_id: str
    user_projects: Union[list, None] = None



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PublicFolder(MongoModel):
    folder_id = StringField(
        max_length=40, generate_id="public-folder", unique=True
    )
    name = StringField(max_length=100)
    tags = DictField(default=None)
    resource_group = StringField(
        max_length=40, choices=("DOMAIN", "WORKSPACE", "PROJECT")
    )
    project_id = StringField(max_length=40)
    workspace_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "tags",
        ],
        "minimal_fields": [
            "folder_id",
            "name",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
        "change_query_keys": {"user_projects": "project_id"},
        "ordering": ["name"],
        "indexes": [
            "name",
            "resource_group",
            "project_id",
            "workspace_id",
            "domain_id",
        ],
    }



import logging
from typing import Tuple
from mongoengine import QuerySet

from spaceone.core.manager import BaseManager
from spaceone.dashboard.model.public_dashboard.database import PublicDashboard

_LOGGER = logging.getLogger(__name__)


class PublicDashboardManager(BaseManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dashboard_model = PublicDashboard

    def create_public_dashboard(self, params: dict) -> PublicDashboard:
        def _rollback(vo: PublicDashboard) -> None:
            _LOGGER.info(
                f"[create_public_dashboard._rollback] "
                f"Delete vo : {vo.name} "
                f"({vo.dashboard_id})"
            )
            vo.delete()

        dashboard_vo: PublicDashboard = self.dashboard_model.create(params)
        self.transaction.add_rollback(_rollback, dashboard_vo)

        return dashboard_vo

    def update_public_dashboard_by_vo(
        self, params: dict, dashboard_vo: PublicDashboard
    ) -> PublicDashboard:
        def _rollback(old_data: dict) -> None:
            _LOGGER.info(
                f"[update_public_dashboard_by_vo._rollback] Revert Data : "
                f'{old_data["dashboard_id"]}'
            )
            dashboard_vo.update(old_data)

        self.transaction.add_rollback(_rollback, dashboard_vo.to_dict())
        return dashboard_vo.update(params)

    @staticmethod
    def delete_public_dashboard_by_vo(dashboard_vo: PublicDashboard) -> None:
        dashboard_vo.delete()

    def get_public_dashboard(
        self,
        dashboard_id: str,
        domain_id: str,
        workspace_id: str = None,
        user_projects=None,
    ) -> PublicDashboard:
        conditions = {
            "dashboard_id": dashboard_id,
            "domain_id": domain_id,
        }

        if workspace_id:
            conditions["workspace_id"] = workspace_id

        if user_projects:
            conditions["project_id"] = user_projects

        return self.dashboard_model.get(**conditions)

    def filter_public_dashboards(self, **conditions) -> QuerySet:
        return self.dashboard_model.filter(**conditions)

    def list_public_dashboards(self, query: dict) -> Tuple[QuerySet, int]:
        return self.dashboard_model.query(**query)

    def stat_public_dashboards(self, query: dict) -> dict:
        return self.dashboard_model.stat(**query)



from spaceone.core.error import *


class ERROR_NOT_SUPPORTED_SOURCE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = "Data table does not support source type. (source_type = {source_type})"


class ERROR_QUERY_OPTION(ERROR_INVALID_ARGUMENT):
    _message = "Query option is invalid. (key = {key})"


class ERROR_NOT_SUPPORTED_QUERY_OPTION(ERROR_INVALID_ARGUMENT):
    _message = "Query option is not supported. (key = {key})"


class ERROR_NOT_SUPPORTED_OPERATOR(ERROR_INVALID_ARGUMENT):
    _message = "Data table does not support operator. (operator = {operator})"



from typing import Union
from pydantic import BaseModel

__all__ = [
    "PrivateFolderCreateRequest",
    "PrivateFolderUpdateRequest",
    "PrivateFolderDeleteRequest",
    "PrivateFolderGetRequest",
    "PrivateFolderSearchQueryRequest",
    "PrivateFolderStatQueryRequest",
]


class PrivateFolderCreateRequest(BaseModel):
    name: str
    tags: Union[dict, None] = None
    dashboards: Union[list, None] = None
    user_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class PrivateFolderUpdateRequest(BaseModel):
    folder_id: str
    name: Union[str, None] = None
    tags: Union[dict, None] = None
    user_id: str
    domain_id: str


class PrivateFolderDeleteRequest(BaseModel):
    folder_id: str
    user_id: str
    domain_id: str


class PrivateFolderGetRequest(BaseModel):
    folder_id: str
    user_id: str
    domain_id: str


class PrivateFolderSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    folder_id: Union[str, None] = None
    name: Union[str, None] = None
    user_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class PrivateFolderStatQueryRequest(BaseModel):
    query: dict
    user_id: str
    domain_id: str



from typing import Union, List, Any
from pydantic import BaseModel

__all__ = [
    "PrivateDashboardCreateRequest",
    "PrivateDashboardUpdateRequest",
    "PrivateDashboardDeleteRequest",
    "PrivateDashboardGetRequest",
    "PrivateDashboardSearchQueryRequest",
    "PrivateDashboardStatQueryRequest",
]


class PrivateDashboardCreateRequest(BaseModel):
    name: str
    description: Union[str, None] = None
    layouts: Union[List[Any], None] = None
    vars: Union[dict, None] = None
    options: Union[dict, None] = None
    variables: Union[dict, None] = None
    variables_schema: Union[dict, None] = None
    labels: Union[List[str], None] = None
    tags: Union[dict, None] = None
    folder_id: Union[str, None] = None
    user_id: str
    workspace_id: Union[str, None] = None
    domain_id: str


class PrivateDashboardUpdateRequest(BaseModel):
    dashboard_id: str
    name: Union[str, None] = None
    description: Union[str, None] = None
    layouts: Union[List[Any], None] = None
    vars: Union[dict, None] = None
    options: Union[dict, None] = None
    variables: Union[dict, None] = None
    variables_schema: Union[dict, None] = None
    labels: Union[List[str], None] = None
    tags: Union[dict, None] = None
    folder_id: Union[str, None] = None
    user_id: str
    domain_id: str


class PrivateDashboardDeleteRequest(BaseModel):
    dashboard_id: str
    user_id: str
    domain_id: str


class PrivateDashboardGetRequest(BaseModel):
    dashboard_id: str
    user_id: str
    domain_id: str


class PrivateDashboardSearchQueryRequest(BaseModel):
    query: Union[dict, None] = None
    dashboard_id: Union[str, None] = None
    name: Union[str, None] = None
    folder_id: Union[str, None] = None
    workspace_id: Union[str, None] = None
    user_id: str
    domain_id: str


class PrivateDashboardStatQueryRequest(BaseModel):
    query: dict
    user_id: str
    domain_id: str



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.dashboard.manager.public_widget_manager import PublicWidgetManager
from spaceone.dashboard.manager.public_dashboard_manager import PublicDashboardManager
from spaceone.dashboard.manager.public_data_table_manager import PublicDataTableManager
from spaceone.dashboard.manager.data_table_manager.data_source_manager import (
    DataSourceManager,
)
from spaceone.dashboard.manager.data_table_manager.data_transformation_manager import (
    DataTransformationManager,
)
from spaceone.dashboard.model.public_widget.request import *
from spaceone.dashboard.model.public_widget.response import *
from spaceone.dashboard.model.public_widget.database import PublicWidget
from spaceone.dashboard.error.dashboard import (
    ERROR_NOT_SUPPORTED_VERSION,
    ERROR_INVALID_PARAMETER,
)

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PublicWidgetService(BaseService):
    resource = "PublicWidget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pub_widget_mgr = PublicWidgetManager()

    @transaction(
        permission="dashboard:PublicWidget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def create(
        self, params: PublicWidgetCreateRequest
    ) -> Union[PublicWidgetResponse, dict]:
        """Create public widget

        Args:
            params (dict): {
                'dashboard_id': 'str',          # required
                'name': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str',             # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicWidgetResponse:
        """

        pub_dashboard_mgr = PublicDashboardManager()
        pub_dashboard_vo = pub_dashboard_mgr.get_public_dashboard(
            params.dashboard_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if pub_dashboard_vo.version == "1.0":
            raise ERROR_NOT_SUPPORTED_VERSION(version=pub_dashboard_vo.version)

        pub_widget_vo = self.pub_widget_mgr.create_public_widget(params.dict())

        return PublicWidgetResponse(**pub_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PublicWidget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def update(
        self, params: PublicWidgetUpdateRequest
    ) -> Union[PublicWidgetResponse, dict]:
        """Update public widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'name': 'str',
                'description': 'str',
                'widget_type': 'str',
                'size': 'str',
                'options': 'dict',
                'data_table_id': 'str',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicWidgetResponse:
        """

        pub_widget_vo: PublicWidget = self.pub_widget_mgr.get_public_widget(
            params.widget_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if params.data_table_id is not None:
            pub_data_table_mgr = PublicDataTableManager()
            pub_data_table_vo = pub_data_table_mgr.get_public_data_table(
                params.data_table_id,
                params.domain_id,
                params.workspace_id,
                params.user_projects,
            )

            if pub_data_table_vo.widget_id != params.widget_id:
                raise ERROR_INVALID_PARAMETER(
                    key="data_table_id",
                    reason="Data table is not belong to this widget.",
                )

        pub_widget_vo = self.pub_widget_mgr.update_public_widget_by_vo(
            params.dict(exclude_unset=True), pub_widget_vo
        )

        return PublicWidgetResponse(**pub_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PublicWidget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def delete(self, params: PublicWidgetDeleteRequest) -> None:
        """Delete public widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_widget_vo: PublicWidget = self.pub_widget_mgr.get_public_widget(
            params.widget_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        self.pub_widget_mgr.delete_public_widget_by_vo(pub_widget_vo)

    @transaction(
        permission="dashboard:PublicWidget.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def load(self, params: PublicWidgetLoadRequest) -> dict:
        """Load public widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'query': 'dict (spaceone.api.core.v1.AnalyzeQuery)', # required
                'vars': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_widget_vo: PublicWidget = self.pub_widget_mgr.get_public_widget(
            params.widget_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if pub_widget_vo.data_table_id is None:
            raise ERROR_INVALID_PARAMETER(
                key="widget_id", reason="Data table is not set."
            )

        pub_data_table_mgr = PublicDataTableManager()
        pub_data_table_vo = pub_data_table_mgr.get_public_data_table(
            pub_widget_vo.data_table_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        if pub_data_table_vo.data_type == "ADDED":
            ds_mgr = DataSourceManager(
                "PUBLIC",
                pub_data_table_vo.source_type,
                pub_data_table_vo.options,
                pub_data_table_vo.widget_id,
                pub_data_table_vo.domain_id,
            )
            return ds_mgr.load_from_widget(
                params.query,
                params.vars,
            )
        else:
            operator = pub_data_table_vo.operator
            options = pub_data_table_vo.options.get(operator, {})

            dt_mgr = DataTransformationManager(
                "PUBLIC",
                pub_data_table_vo.operator,
                options,
                pub_data_table_vo.widget_id,
                pub_data_table_vo.domain_id,
            )
            return dt_mgr.load_from_widget(
                params.query,
                params.vars,
            )

    @transaction(
        permission="dashboard:PublicWidget.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def get(self, params: PublicWidgetGetRequest) -> Union[PublicWidgetResponse, dict]:
        """Get public widget

        Args:
            params (dict): {
                'widget_id': 'str',             # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicWidgetResponse:
        """

        pub_widget_vo: PublicWidget = self.pub_widget_mgr.get_public_widget(
            params.widget_id,
            params.domain_id,
            params.workspace_id,
            params.user_projects,
        )

        return PublicWidgetResponse(**pub_widget_vo.to_dict())

    @transaction(
        permission="dashboard:PublicWidget.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(
        [
            "dashboard_id",
            "widget_id",
            "name",
            "domain_id",
            "workspace_id",
            "project_id",
            "user_projects",
        ]
    )
    @append_keyword_filter(["widget_id", "name"])
    @convert_model
    def list(
        self, params: PublicWidgetSearchQueryRequest
    ) -> Union[PublicWidgetsResponse, dict]:
        """List public widgets

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'dashboard_id': 'str',                          # required
                'widget_id': 'str',
                'name': 'str',
                'project_id': 'str',
                'workspace_id': 'str',                          # injected from auth
                'domain_id': 'str',                             # injected from auth (required)
                'user_projects': 'list',                        # injected from auth
            }

        Returns:
            PublicWidgetsResponse:
        """

        query = params.query or {}
        pub_widget_vos, total_count = self.pub_widget_mgr.list_public_widgets(query)
        pub_widgets_info = [pub_widget_vo.to_dict() for pub_widget_vo in pub_widget_vos]
        return PublicWidgetsResponse(results=pub_widgets_info, total_count=total_count)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PrivateWidget(MongoModel):
    widget_id = StringField(max_length=40, generate_id="private-widget", unique=True)
    name = StringField(max_length=100)
    description = StringField(default=None)
    widget_type = StringField(max_length=40, default="NONE")
    size = StringField(default=None, null=True)
    options = DictField(default=None, null=True)
    tags = DictField(default=None)
    data_table_id = StringField(max_length=40, default=None, null=True)
    dashboard_id = StringField(max_length=40)
    user_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        "updatable_fields": [
            "name",
            "description",
            "widget_type",
            "size",
            "options",
            "tags",
            "data_table_id",
        ],
        "minimal_fields": [
            "widget_id",
            "name",
            "widget_type",
            "data_table_id",
            "dashboard_id",
            "user_id",
            "domain_id",
        ],
        "ordering": ["name"],
        "indexes": [
            "name",
            "widget_type",
            "data_table_id",
            "dashboard_id",
            "user_id",
            "domain_id",
        ],
    }



import logging
from typing import Union

from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.dashboard.manager.public_folder_manager import PublicFolderManager
from spaceone.dashboard.manager.identity_manager import IdentityManager
from spaceone.dashboard.model.public_folder.request import *
from spaceone.dashboard.model.public_folder.response import *
from spaceone.dashboard.model.public_folder.database import PublicFolder

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class PublicFolderService(BaseService):
    resource = "PublicFolder"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pub_folder_mgr = PublicFolderManager()
        self.identity_mgr = IdentityManager()

    @transaction(
        permission="dashboard:PublicFolder.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def create(self, params: PublicFolderCreateRequest) -> Union[PublicFolderResponse, dict]:
        """Create public folder

        Args:
            params (dict): {
                'name': 'str',                  # required
                'tags': 'dict',
                'dashboards': 'list',
                'resource_group': 'str',        # required
                'project_id': 'str',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
            }

        Returns:
            PublicFolderResponse:
        """

        if params.resource_group == "PROJECT":
            if not params.project_id:
                raise ERROR_REQUIRED_PARAMETER(key="project_id")

            project_info = self.identity_mgr.get_project(params.project_id)
            params.workspace_id = project_info["workspace_id"]
        elif params.resource_group == "WORKSPACE":
            if not params.workspace_id:
                raise ERROR_REQUIRED_PARAMETER(key="workspace_id")

            self.identity_mgr.check_workspace(params.workspace_id, params.domain_id)
            params.project_id = "*"
        else:
            params.workspace_id = "*"
            params.project_id = "*"

        pub_folder_vo = self.pub_folder_mgr.create_public_folder(params.dict())

        # Create child dashboards
        if params.dashboards:
            pass

        return PublicFolderResponse(**pub_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PublicFolder.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def update(self, params: PublicFolderUpdateRequest) -> Union[PublicFolderResponse, dict]:
        """Update public folder

        Args:
            params (dict): {
                'folder_id': 'str',             # required
                'name': 'str',
                'tags': 'dict',
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            PublicFolderResponse:
        """

        pub_folder_vo: PublicFolder = self.pub_folder_mgr.get_public_folder(
            params.folder_id, params.domain_id, params.workspace_id, params.user_projects
        )

        pub_folder_vo = self.pub_folder_mgr.update_public_folder_by_vo(
            params.dict(exclude_unset=True), pub_folder_vo)

        return PublicFolderResponse(**pub_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PublicFolder.write",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def delete(self, params: PublicFolderDeleteRequest) -> None:
        """Delete public folder

        Args:
            params (dict): {
                'folder_id': 'str',             # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list'         # injected from auth
            }

        Returns:
            None
        """

        pub_folder_vo: PublicFolder = self.pub_folder_mgr.get_public_folder(
            params.folder_id, params.domain_id, params.workspace_id, params.user_projects
        )

        self.pub_folder_mgr.delete_public_folder_by_vo(pub_folder_vo)

    @transaction(
        permission="dashboard:PublicFolder.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @convert_model
    def get(self, params: PublicFolderGetRequest) -> Union[PublicFolderResponse, dict]:
        """Get public folder

        Args:
            params (dict): {
                'folder_id': 'str',             # required
                'workspace_id': 'str',          # injected from auth
                'domain_id': 'str'              # injected from auth (required)
                'user_projects': 'list',        # injected from auth
            }

        Returns:
            PublicFolderResponse:
        """

        pub_folder_vo: PublicFolder = self.pub_folder_mgr.get_public_folder(
            params.folder_id, params.domain_id, params.workspace_id, params.user_projects
        )

        return PublicFolderResponse(**pub_folder_vo.to_dict())

    @transaction(
        permission="dashboard:PublicFolder.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(
        ["folder_id", "name", "domain_id", "workspace_id", "project_id", "user_projects"]
    )
    @append_keyword_filter(["folder_id", "name"])
    @convert_model
    def list(self, params: PublicFolderSearchQueryRequest) -> Union[PublicFoldersResponse, dict]:
        """List public folders

        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.Query)'
                'folder_id': 'str',
                'name': 'str',
                'project_id': 'str',
                'workspace_id': 'str',                          # injected from auth
                'domain_id': 'str',                             # injected from auth (required)
                'user_projects': 'list',                        # injected from auth
            }

        Returns:
            PublicFoldersResponse:
        """

        query = params.query or {}
        pub_folder_vos, total_count = self.pub_folder_mgr.list_public_folders(query)
        pub_dashboards_info = [pub_folder_vo.to_dict() for pub_folder_vo in pub_folder_vos]
        return PublicFoldersResponse(results=pub_dashboards_info, total_count=total_count)

    @transaction(
        permission="dashboard:PublicFolder.read",
        role_types=["DOMAIN_ADMIN", "WORKSPACE_OWNER", "WORKSPACE_MEMBER"],
    )
    @append_query_filter(["domain_id", "workspace_id", "user_projects"])
    @append_keyword_filter(["folder_id", "name"])
    @convert_model
    def stat(self, params: PublicFolderStatQueryRequest) -> dict:
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
                'workspace_id': 'str',                                  # injected from auth
                'domain_id': 'str'                                      # injected from auth (required)
                'user_projects': 'list',                                # injected from auth
            }

        Returns:
            dict:

        """

        query = params.query or {}
        return self.pub_folder_mgr.stat_public_folders(query)
